@@ -33,6 +33,7 @@ module Data.GraphViz.Parsing
     , runParserWith
     , parseLiberally
     , checkValidParse
+    , checkValidParseWithRest
       -- * Convenience parsing combinators.
     , ignoreSep
     , onlyBool
@@ -77,7 +78,7 @@ module Data.GraphViz.Parsing
     , parseColorScheme
     ) where
 
-import Data.GraphViz.Exception      (GraphvizException (NotDotCode), throw)
+import Data.GraphViz.Exception      (GraphvizException(NotDotCode), throw)
 import Data.GraphViz.Internal.State
 import Data.GraphViz.Internal.Util
 
@@ -98,10 +99,11 @@ import           Data.Maybe          (fromMaybe, isJust, isNothing, listToMaybe,
                                       maybeToList)
 import           Data.Ratio          ((%))
 import qualified Data.Set            as Set
+import qualified Data.Text           as ST
 import           Data.Text.Lazy      (Text)
 import qualified Data.Text.Lazy      as T
 import qualified Data.Text.Lazy.Read as T
-import           Data.Version        (Version (..))
+import           Data.Version        (Version(..))
 import           Data.Word           (Word16, Word8)
 
 -- -----------------------------------------------------------------------------
@@ -125,7 +127,7 @@ runParserWith f p t = let (r,_,t') = P.runParser p (f initialState) t
 --   parsing function consumes all of the 'Text' input (with the
 --   exception of whitespace at the end).
 runParser'   :: Parse a -> Text -> a
-runParser' p = checkValidParse . fst . runParser p'
+runParser' p = checkValidParseWithRest . runParser p'
   where
     p' = p `discard` (whitespace *> eof)
 
@@ -158,6 +160,15 @@ checkValidParse :: Either String a -> a
 checkValidParse (Left err) = throw (NotDotCode err)
 checkValidParse (Right a)  = a
 
+-- | If unable to parse /Dot/ code properly, 'throw' a
+--   'GraphvizException', with the error containing the remaining
+--   unparsed code..
+checkValidParseWithRest :: (Either String a, Text) -> a
+checkValidParseWithRest (Left err, rst) = throw (NotDotCode err')
+  where
+    err' = err ++ "\n\nRemaining input:\n\t" ++ show rst
+checkValidParseWithRest (Right a,_)     = a
+
 -- | Parse the required value with the assumption that it will parse
 --   all of the input 'Text'.
 parseIt' :: (ParseDot a) => Text -> a
@@ -239,6 +250,11 @@ instance ParseDot Text where
           -- above.
           quotedParse quotedString
 
+instance ParseDot ST.Text where
+  parseUnqt = T.toStrict <$> parseUnqt
+
+  parse = T.toStrict <$> parse
+
 instance (ParseDot a) => ParseDot [a] where
   parseUnqt = parseUnqtList
 
