@@ -18,10 +18,9 @@
  -}
 module Data.GraphViz.Testing.Instances() where
 
-import Data.GraphViz.Testing.Instances.FGL()
-import Data.GraphViz.Testing.Instances.Canonical()
-import Data.GraphViz.Testing.Instances.Generalised()
-import Data.GraphViz.Testing.Instances.Graph()
+import Data.Graph.Inductive.Arbitrary              ()
+import Data.GraphViz.Testing.Instances.Canonical   ()
+import Data.GraphViz.Testing.Instances.Generalised ()
+import Data.GraphViz.Testing.Instances.Graph       ()
 
 -- -----------------------------------------------------------------------------
-
